# Generated by Django 2.2.24 on 2021-10-11 18:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("core", "0013_auto_20210711_1105")]

    operations = [
        migrations.RenameField(
            model_name="requestlog", old_name="address", new_name="session_key"
        )
    ]
