# Generated by Django 2.2.24 on 2021-11-28 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("core", "0015_install_trgm_extension")]

    operations = [
        migrations.AddField(
            model_name="archivedsong",
            name="cached",
            field=models.BooleanField(default=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="archivedsong",
            name="duration",
            # with a default of -1,
            # existing uncached songs (e.g. spotify) will have an unknown duration.
            # To fix this, run a one-off script that fetches the duration for all old uncached songs
            # New songs are not affected, they will have the correct duration.
            field=models.FloatField(default=-1),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="currentsong", name="duration", field=models.FloatField()
        ),
        migrations.AlterField(
            model_name="queuedsong", name="duration", field=models.FloatField()
        ),
    ]
