# Generated by Django 4.1.7 on 2023-05-05 19:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0017_archivedquery_core_archivedquery_query_trgm_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="setting",
            name="value",
            field=models.CharField(max_length=1000),
        ),
    ]
